/**************************************************************************
***
***    Copyright (c) 1995-1998 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this package may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

#include "1balanceGen.h"
#include "hgFixed.h"
#include "stats.h"

// SBGWeightRand: implemented by ILM
// produces a simple-minded randomized partitioning which is balanced
// with high probability; FM, Sanchis, SA and other consumers will need
// to make improving passes -- they can combine such passes with improving
// their objectives, which is why this methods does not do additional improving
// passes (this can be added as an option if needed)
//
// Possible improvements:
//  1. some "local" random permuting needed for nodes of approx. equal sizes
//   because the locations of consequtive nodes follownig a node with
//   much bigger weight can be highly correlated
//   (permuting all nodes is not good since it can decrease the probability
//   of getting a balances partition if there very big nodes)
//  2. Can we estimate the probability of getting a balanced solution ?
//     How do we improve the probability

 
vector<double> SBGWeightRand::generateSoln(Partitioning& curPart)
{
    unsigned numPartitions = _problem.getNumPartitions();
    const HGraphFixed& hgraph = _problem.getHGraph();

    abkfatal(hgraph.getNumNodes() >= numPartitions, 
             "not enough nodes / partition");
    
    vector<double> areaToMin    (numPartitions),
                   areaToFill   (numPartitions),
		   targetAreas  (numPartitions),
                   availableArea(numPartitions),
                   chancesOfGoingToPart(numPartitions);

    const vector<vector<double> >& maxCaps = _problem.getMaxCapacities();
    const vector<vector<double> >& minCaps = _problem.getMinCapacities();
    const vector<vector<double> >& caps =    _problem.getCapacities();

    for(unsigned k = 0; k < numPartitions; k++)
    {
        availableArea[k] = maxCaps[k][0];
        areaToMin[k]     = minCaps[k][0];
        areaToFill[k]    = caps[k][0];
        targetAreas[k]   = caps[k][0];
    }

    unsigned m, numMovables = hgraph.getNumNodes();
 
    const Permutation& sortAscWts_1=hgraph.getNodesSortedByWeights();
    RandomRawUnsigned ru;
    
    for(m=numMovables-1 ; m != (-1); m--)
    {
        unsigned nodeIdx= sortAscWts_1[m];
        HGFNode& node=hgraph.getNodeByIdx(nodeIdx);
        double   nodeWt = node.getWeight();
        double   minNonZero=DBL_MAX;
        unsigned nonZeroChances=0; // to how many partitions a node can go
  
        if (nodeWt==0.0)
        {
           for(unsigned p = 0; p != numPartitions; p++)
             if (curPart[nodeIdx].isInPart(p)) nonZeroChances++;
           if (nonZeroChances==0)
           {
              cerr << " Node " << nodeIdx << " is constrainted to "
                   << curPart[nodeIdx];
              abkfatal(0," Node can't go to any partition ");
           }
           k=ru % nonZeroChances;
           nonZeroChances=0;
           for(p=0; p != numPartitions; p++)
           {
              if (curPart[nodeIdx].isInPart(p))
                 if (nonZeroChances++ == k) { k=p;  goto setpart; }
           }
           abkfatal(0," Internal error: control should not reach here");
        }
 
        // compute chancesOfGoingToPart[] assuming some partitions 
        // are still underfilled
        for(k = 0; k < numPartitions; k++)
        {
            if ( curPart[nodeIdx].isInPart(k) && nodeWt <= areaToMin[k] + 1e-4)
            {
                nonZeroChances++;
                chancesOfGoingToPart[k] = areaToMin[k];
                if (chancesOfGoingToPart[k] < minNonZero)
                   minNonZero=chancesOfGoingToPart[k];
            }
            else  chancesOfGoingToPart[k] = 0.0;
        }

        // If chances of going to some partition are nonZero ...

        if (nonZeroChances) goto selection;

        // otherwise (e.g. all partitions are filled at least to minimum size),
        // recompute chancesOfGoingToPart[]
        // assuming some partitions are below their targets
 
        for(k = 0; k < numPartitions; k++)
        {
            if ( curPart[nodeIdx].isInPart(k) && nodeWt <= areaToFill[k] + 1e-4)            {  
                nonZeroChances++;
                chancesOfGoingToPart[k] = areaToFill[k];
                if (chancesOfGoingToPart[k] < minNonZero)
                   minNonZero=chancesOfGoingToPart[k];
            }
            else  chancesOfGoingToPart[k] = 0;
        }

        // If chances of going to some partition are nonZero ...

        if (nonZeroChances) goto selection;

        // otherwise (e.g. all partitions are filled at least to target size),
        // recompute chancesOfGoingToPart[]
        // assuming some partitions are below their max
 
        for(k = 0; k < numPartitions; k++)
        {
          if ( curPart[nodeIdx].isInPart(k) && nodeWt <= availableArea[k] +1e-4)          {  
              nonZeroChances++;
              chancesOfGoingToPart[k] = availableArea[k];
              if (chancesOfGoingToPart[k] < minNonZero)
                 minNonZero=chancesOfGoingToPart[k];
          }
          else  chancesOfGoingToPart[k] = 0;
        }
 
        // If chances of going to some partition are nonZero ...

        if (nonZeroChances) goto selection;

        // otherwise (e.g. all partitions are filled at least to max size),
        // recompute chancesOfGoingToPart[] to minimizing overfill
 
        for(k = 0; k < numPartitions; k++)
        {
          if ( curPart[nodeIdx].isInPart(k) )
          {
             nonZeroChances++;
             chancesOfGoingToPart[k] = 1.0 / abs( nodeWt - availableArea[k] );
             if (chancesOfGoingToPart[k] < minNonZero)
                 minNonZero=chancesOfGoingToPart[k];
          }
          else chancesOfGoingToPart[k] = 0;
        }

        // if chance of going to all partitions are 0 i
        // (e.g. the node was not allowed anywhere), issue fatal error messg.

        abkfatal3(nonZeroChances,
               " Node ", nodeIdx, " can't go into any partition ");
 
 selection : minNonZero *= 1e-3;

        if (nonZeroChances==1) 
        {
            for(k = 0; k < numPartitions; k++)
            if (chancesOfGoingToPart[k]>0.0) goto setpart;
            abkfatal(0," Internal error: control should not reach here");
        }
        else
        {
            for(k = 0; k < numPartitions; k++)
            chancesOfGoingToPart[k] /= minNonZero;
            k = BiasedRandomSelection(chancesOfGoingToPart, _ru);
        }
 
setpart: curPart[nodeIdx].setToPart(k);
        areaToMin    [k] -= nodeWt;
        areaToFill   [k] -= nodeWt;
        availableArea[k] -= nodeWt;
    }

    vector<double> partSizes(numPartitions);
    for(k = 0; k < numPartitions; k++)
        partSizes[k]= targetAreas[k] - areaToFill[k];
 
    return partSizes;
}

vector<double> SBGWeightRandDutt::generateSoln(Partitioning& curPart)
{
    unsigned numPartitions    = _problem.getNumPartitions();
    const HGraphFixed& hgraph = _problem.getHGraph();

    abkfatal(numPartitions==2,
     " This ini soln generator is only good for 2 parts");

    abkfatal(hgraph.getNumNodes() >= numPartitions, 
             "not enough nodes / partition");

//  cout << " Ini solution :" << curPart << endl;

    const vector<vector<double> >& maxCaps = _problem.getMaxCapacities();
    const vector<vector<double> >& targets = _problem.getCapacities();
    vector<double> availableArea(numPartitions);

  //for(unsigned k = 0; k < numPartitions; k++) availableArea[k]=maxCaps[k][0];
    for(unsigned k = 0; k < numPartitions; k++) availableArea[k]=targets[k][0];

    unsigned m, numMovables = hgraph.getNumNodes();
 
    const Permutation& sortAscWts_1=hgraph.getNodesSortedByWeights();
    double minCellSize=hgraph.getNodeByIdx(0).getWeight();

    RandomRawUnsigned ru;
    PartitionIds freeToMove, cantMove;
    freeToMove.setToAll(2);
    
    for(m=numMovables-1 ; m != (-1); m--)
    {
        unsigned nodeIdx= sortAscWts_1[m];
        //unsigned nodeIdx=m;

        abkfatal3 (curPart[nodeIdx] != cantMove, "Node ", nodeIdx,
                   " cannot be assigned to a partition\n");

        double   nodeWt = hgraph.getNodeByIdx(nodeIdx).getWeight();

        if (curPart[nodeIdx] == freeToMove)
        { 
           unsigned side= (ru % 2 ? 1 : 0);
           if (availableArea[side]>minCellSize+nodeWt)
           {
               curPart[nodeIdx].setToPart(side);
               availableArea[side]-=nodeWt;
           }
	   else
           {
              curPart[nodeIdx].setToPart(1-side);
             availableArea[1-side]-=nodeWt;
           }
        }
        else
        {
           unsigned side= (curPart[nodeIdx].isInPart(0) ? 0 : 1);
           availableArea[side]-=nodeWt;
        }
    }

    vector<double> partSizes(numPartitions);
    for(k = 0; k < numPartitions; k++) 
       partSizes[k]=maxCaps[k][0]-availableArea[k];
       //partSizes[k]=targets[k][0]-availableArea[k];

    return partSizes;
} 

vector<double> SBGWeightRandLC::generateSoln(Partitioning& curPart )
{
    unsigned numPartitions = _problem.getNumPartitions();
    const HGraphFixed& hgraph = _problem.getHGraph();

    abkfatal(hgraph.getNumNodes() >= numPartitions,
          "not enough nodes / partition");
    
    vector<double> areaToMin    (numPartitions),
                   areaToFill   (numPartitions),
		   targetAreas  (numPartitions),
                   availableArea(numPartitions),
                   chancesOfGoingToPart(numPartitions);

    const vector<vector<double> >& maxCaps = _problem.getMaxCapacities();
    const vector<vector<double> >& minCaps = _problem.getMinCapacities();
    const vector<vector<double> >& caps =    _problem.getCapacities();

    double minAbsTol=DBL_MAX;
    for(unsigned k = 0; k < numPartitions; k++)
    {
        availableArea[k] = maxCaps[k][0];
        areaToMin[k]     = minCaps[k][0];
        areaToFill[k]    = caps[k][0];
        targetAreas[k]   = caps[k][0];
        double absTol=availableArea[k] - areaToMin[k];
        if (absTol<minAbsTol) minAbsTol=absTol;
    }

    unsigned m, numMovables = hgraph.getNumNodes();
    
    const Permutation& sortAscWts_1=hgraph.getNodesSortedByWeights();
    RandomRawUnsigned ru;

    // This is what's different between WeightRand and WeightRandLC generators

    // assign at random for a while
    for(m=numMovables-1;  m != (-1); m--)
    {
        unsigned nodeIdx= sortAscWts_1[m];
        HGFNode& node=hgraph.getNodeByIdx(nodeIdx);
        double   nodeWt = node.getWeight();
//      cout << " nodeWt " << nodeWt << " minAbsTol " << minAbsTol << endl;

        if (2*nodeWt<=minAbsTol) 
        { 
//        cout << "*Switched at m="<<numMovables-1-m<<endl;
          break;
        }

        unsigned nonZeroChances=0; // to how many partitions a node can go
  
//      if (nodeWt==0.0)
        {
           for(unsigned p = 0; p != numPartitions; p++)
             if (curPart[nodeIdx].isInPart(p)) nonZeroChances++;
           if (nonZeroChances==0)
           {
              cerr << " Node " << nodeIdx << " is constrainted to "
                   << curPart[nodeIdx];
              abkfatal(0," Node can't go to any partition ");
           }
           k=ru % nonZeroChances;
           nonZeroChances=0;
           for(p=0; p != numPartitions; p++)
           {
              if (curPart[nodeIdx].isInPart(p))
                 if (nonZeroChances++ == k) { k=p;  goto setpart1; }
           }
           abkfatal(0," Internal error: control should not reach here");
        }

setpart1: 
        curPart[nodeIdx].setToPart(k);
        areaToMin    [k] -= nodeWt;
        areaToFill   [k] -= nodeWt;
        availableArea[k] -= nodeWt;
        if (targetAreas[k]>=2*areaToFill[k]) 
        { 
          m--;
//        cout << " Switched at m="<<numMovables-1-m<<endl;
          break;
        }
    }
    // end difference

/*
    for(; m != (-1); m--)
    {
        unsigned nodeIdx= sortAscWts_1[m];
        HGFNode& node=hgraph.getNodeByIdx(nodeIdx);
        double   nodeWt = node.getWeight();
        double   minNonZero=DBL_MAX;
        unsigned nonZeroChances=0; // to how many partitions a node can go
  
        if (nodeWt==0.0)
        {
           for(unsigned p = 0; p != numPartitions; p++)
             if (curPart[nodeIdx].isInPart(p)) nonZeroChances++;
           if (nonZeroChances==0)
           {
              cerr << " Node " << nodeIdx << " is constrainted to "
                   << curPart[nodeIdx];
              abkfatal(0," Node can't go to any partition ");
           }
           k=ru % nonZeroChances;
           nonZeroChances=0;
           for(p=0; p != numPartitions; p++)
           {
              if (curPart[nodeIdx].isInPart(p))
                 if (nonZeroChances++ == k) { k=p;  goto setpart; }
           }
           abkfatal(0," Internal error: control should not reach here");
        }
 
        // compute chancesOfGoingToPart[] assuming some partitions 
        // are still underfilled
        for(k = 0; k < numPartitions; k++)
        {
            if ( curPart[nodeIdx].isInPart(k) && nodeWt <= areaToMin[k] + 1e-4)
            {
                nonZeroChances++;
                chancesOfGoingToPart[k] = areaToMin[k];
                if (chancesOfGoingToPart[k] < minNonZero)
                   minNonZero=chancesOfGoingToPart[k];
            }
            else  chancesOfGoingToPart[k] = 0.0;
        }

        // If chances of going to some partition are nonZero ...

        if (nonZeroChances) goto selection;

        // otherwise (e.g. all partitions are filled at least to minimum size),
        // recompute chancesOfGoingToPart[]
        // assuming some partitions are below their targets
 
        for(k = 0; k < numPartitions; k++)
        {
            if ( curPart[nodeIdx].isInPart(k) && nodeWt <= areaToFill[k] + 1e-4)            {  
                nonZeroChances++;
                chancesOfGoingToPart[k] = areaToFill[k];
                if (chancesOfGoingToPart[k] < minNonZero)
                   minNonZero=chancesOfGoingToPart[k];
            }
            else  chancesOfGoingToPart[k] = 0;
        }

        // If chances of going to some partition are nonZero ...

        if (nonZeroChances) goto selection;

        // otherwise (e.g. all partitions are filled at least to target size),
        // recompute chancesOfGoingToPart[]
        // assuming some partitions are below their max
 
        for(k = 0; k < numPartitions; k++)
        {
          if ( curPart[nodeIdx].isInPart(k) && nodeWt <= availableArea[k] +1e-4)          {  
              nonZeroChances++;
              chancesOfGoingToPart[k] = availableArea[k];
              if (chancesOfGoingToPart[k] < minNonZero)
                 minNonZero=chancesOfGoingToPart[k];
          }
          else  chancesOfGoingToPart[k] = 0;
        }
 
        // If chances of going to some partition are nonZero ...

        if (nonZeroChances) goto selection;

        // otherwise (e.g. all partitions are filled at least to max size),
        // recompute chancesOfGoingToPart[] to minimizing overfill
 
        for(k = 0; k < numPartitions; k++)
        {
          if ( curPart[nodeIdx].isInPart(k) )
          {
             nonZeroChances++;
             chancesOfGoingToPart[k] = 1.0 / abs( nodeWt - availableArea[k] );
             if (chancesOfGoingToPart[k] < minNonZero)
                 minNonZero=chancesOfGoingToPart[k];
          }
          else chancesOfGoingToPart[k] = 0;
        }

        // if chance of going to all partitions are 0 i
        // (e.g. the node was not allowed anywhere), issue fatal error messg.

        abkfatal3(nonZeroChances,
               " Node ", nodeIdx, " can't go into any partition ");
 
 selection : minNonZero *= 1e-3;

        if (nonZeroChances==1) 
        {
            for(k = 0; k < numPartitions; k++)
            if (chancesOfGoingToPart[k]>0.0) goto setpart;
            abkfatal(0," Internal error: control should not reach here");
        }
        else
        {
            for(k = 0; k < numPartitions; k++)
            chancesOfGoingToPart[k] /= minNonZero;
            k = BiasedRandomSelection(chancesOfGoingToPart, _ru);
        }
 
setpart: curPart[nodeIdx].setToPart(k);
        areaToMin    [k] -= nodeWt;
        areaToFill   [k] -= nodeWt;
        availableArea[k] -= nodeWt;
    }

    vector<double> partSizes(numPartitions);
    for(k = 0; k < numPartitions; k++)
        partSizes[k]= targetAreas[k] - areaToFill[k];
 
    return partSizes;
*/

  SBGWeightRand forRemainingCells(_problem);
  return forRemainingCells.generateSoln(curPart);
}

vector<double> 
SBGWeightRandLCwBFS::generateSoln(Partitioning& curPart )
{
    unsigned numPartitions = _problem.getNumPartitions();
    const HGraphFixed& hgraph = _problem.getHGraph();

    abkfatal(numPartitions==2,
     " This ini soln generator is only good for 2 parts");

    abkfatal(hgraph.getNumNodes() >= numPartitions,
          "not enough nodes / partition");
    
    vector<double> areaToMin    (numPartitions),
                   areaToFill   (numPartitions),
		   targetAreas  (numPartitions),
                   availableArea(numPartitions),
                   chancesOfGoingToPart(numPartitions);

    const vector<vector<double> >& maxCaps = _problem.getMaxCapacities();
    const vector<vector<double> >& minCaps = _problem.getMinCapacities();
    const vector<vector<double> >& caps =    _problem.getCapacities();

    double minAbsTol=DBL_MAX;
    for(unsigned k = 0; k < numPartitions; k++)
    {
        availableArea[k] = maxCaps[k][0];
        areaToMin[k]     = minCaps[k][0];
        areaToFill[k]    = caps[k][0];
        targetAreas[k]   = caps[k][0];
        double absTol=availableArea[k] - areaToMin[k];
        if (absTol<minAbsTol) minAbsTol=absTol;
    }
    double minTarget=min(targetAreas[0],targetAreas[1]);

    unsigned m, numMovables = hgraph.getNumNodes();
    
    const Permutation& sortAscWts_1=hgraph.getNodesSortedByWeights();
    RandomRawUnsigned ru;

    // find large cells indices
    vector<unsigned>  largeCells; largeCells.reserve(20);
    double totalAreaOfLargeCells=0.0;
   
    cout << endl;
    for(m=numMovables-1;  m != (-1); m--)
    {
        unsigned nodeIdx= sortAscWts_1[m];
        double   nodeWt = hgraph.getNodeByIdx(nodeIdx).getWeight();
//      cout << " Cell " << nodeIdx << " weight " << nodeWt 
//           << " minAbsTol " << minAbsTol << " minTarget " << minTarget 
//           << endl;
        if (2*nodeWt<=minAbsTol) break;
        largeCells.push_back(nodeIdx);
        totalAreaOfLargeCells+=nodeWt;

        if (2*totalAreaOfLargeCells>minTarget) 
        { 
          // cout << "Switched on total capacity " << endl; 
          break;
        }
    }
//  cout << " Found " << largeCells.size() << " large cells : \n"
//       << largeCells << endl;

  if (largeCells.size()) 
  {
    vector<unsigned> fixedInPart0,fixedInPart1;
    for(m=numMovables-1;  m != (-1); m--)
    { 
        if (!curPart[m].isInPart(0)) fixedInPart1.push_back(m);
        else
        if (!curPart[m].isInPart(1)) fixedInPart0.push_back(m);
    }

    vector<unsigned> toProcess, startPart, curNeighb;
    toProcess.reserve(numMovables);
    startPart .reserve(numMovables);
    curNeighb.reserve(numMovables);

    bit_vector visited(numMovables,false);
    bool startInPart1=false;
    double visitedArea=0.0;
    if (!fixedInPart0.empty())
    {
//    cout << " Starting BFS with fixed in part 0" << endl;
      curNeighb=startPart=fixedInPart0;
      for(k=0;k!=startPart.size(); k++)
      {
        unsigned idx=startPart[k];
        visitedArea+=hgraph.getNodeByIdx(idx).getWeight(); 
        visited[idx]=true;
      }
    }
    else
    if (!fixedInPart1.empty())
    {
      startInPart1=true;
//    cout << " Starting BFS with fixed in part 1" << endl;
      curNeighb=startPart=fixedInPart0;
      for(k=0;k!=startPart.size(); k++)
      {
        unsigned idx=startPart[k];
        visitedArea+=hgraph.getNodeByIdx(idx).getWeight(); 
        visited[idx]=true;
      }
    }
    else
    { 
       // choose a largeCell at random 
       unsigned firstVisited=largeCells[ru % largeCells.size()];
//     cout << " No fixed cells: starting BFS with cell " 
//          << firstVisited << " in part 0" << endl;
       curNeighb.push_back(firstVisited);
       startPart .push_back(firstVisited);
       visitedArea=hgraph.getNodeByIdx(firstVisited).getWeight();
       visited[firstVisited]=true;
    }

    // do randomized BFS 
    unsigned neighbSize=curNeighb.size();
    RandomUnsigned ru2(0,10);
   
    unsigned round=0;

    while( neighbSize )
    {
      toProcess=curNeighb;
      random_shuffle(toProcess.begin(), toProcess.end(), ru2);
      curNeighb.clear();

//    cout << "BFS round " << ++round 
//         << " to process : " << neighbSize << endl;

      for(k=0; k!=neighbSize; k++)
      {
	 unsigned idx=toProcess.back();
	 toProcess.pop_back();
	 HGFNode& node=hgraph.getNodeByIdx(idx); 
	 for(itHGFEdgeLocal edgeIt =node.edgesBegin();
	     edgeIt!=node.edgesEnd(); edgeIt++)
	 {
	  HGFEdge& edge=(**edgeIt);
	  if (edge.getDegree()>3) continue;
	  itHGFNodeLocal nodeIt=edge.nodesBegin();
	  for(; nodeIt!=edge.nodesEnd(); nodeIt++)
	  {
	     unsigned nodeIdx=(*nodeIt)->getIndex();
	     if (nodeIdx==idx) continue;
	     if (visited[nodeIdx]) continue;
	     visited[nodeIdx]=true;
	     visitedArea+=hgraph.getNodeByIdx(nodeIdx).getWeight();
	     if (visitedArea>targetAreas[0]) goto assignLargeCells; 
	     curNeighb.push_back(nodeIdx);
	     startPart .push_back(nodeIdx); 
	     
	  }
	 }
      } 
      neighbSize=curNeighb.size();
    }

 assignLargeCells: 
//  cout << " BFS visited " << startPart.size() << " cells out of "
//<< numMovables << endl;
    for(k=0; k!=largeCells.size(); k++)
    {
       // possibly overriden below
       curPart[largeCells[k]].setToPart( startInPart1 ? 0 : 1);
    }

    sort(startPart.begin(), startPart.end()); 
    sort(largeCells.begin(),largeCells.end());

    vector<unsigned> largeCellsInStartPart(largeCells);
    set_intersection(startPart.begin(), startPart.end(), 
		     largeCells.begin(),largeCells.end(),
		     largeCellsInStartPart.begin());
//  cout << " Large cells in starting part :\n" 
//       << largeCellsInStartPart << endl;
    
    for(k=0; k!=largeCellsInStartPart.size(); k++)
       curPart[largeCellsInStartPart[k]].setToPart( startInPart1 ? 1 : 0); 
  }

    SBGWeightRand forRemainingCells(_problem);
    return forRemainingCells.generateSoln(curPart);
}
