/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/







//Base Small Placer
// - Small Placement Problem
//   created by Andrew Caldwell  981206

#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include "sRowDPPlacer.h"
#include "stripNetStacks.h"

DPSmallPlacer::DPSmallPlacer(SmallPlacementProblem& problem,
                               Parameters params)                          
	: BaseSmallPlacer(problem, params), 
	  _curSoln(problem.getNumCells()),
	  _bestSolnOverPass(problem.getNumCells())
{
    abkfatal(problem.getNumRows() == 1,
		"DP SmallPlacer can't deal with multiple rows") ;
    abkfatal(problem.getNumCells() >= 2,
		"DP SmallPlacer requires at least 2 cells") ;

    unsigned bits_in_uns = sizeof(unsigned)*8 ;
    abkfatal(problem.getNumNets() < 2 * bits_in_uns,
		"DP SmallPlacer can't deal with this many nets") ;

    Placement& soln    = _problem.getSolnBuffer()[0].placement ;
//  const unsigned numNets  = _problem.getNumNets() ;
    const unsigned numCells = _problem.getNumCells() ;

    buildConnectivity() ;

        // set y-coordinates (single row, so these don't change!)
    unsigned n ;
    for(n = numCells ; n > 0 ; ) {
        --n ;
        _curSoln[n].y = soln[n].y = problem.getRows()[0].yMin ;
    }

    // for the moment, split the problem into two equal parts.

    bool usingInitSoln = false ;

    _q.reserve(numCells) ; // workspace to get init soln.

    if(_problem.getSolnBuffer()[0].populated && _params.useInitSoln)
    {
        for (n = 0 ; n < numCells ; n++) 
            _q.push_back(n) ;

        std::sort(_q.begin(), _q.end(), DPCompareNodesByInitPl(soln));
        usingInitSoln = true;
    }
    else {
        if (_params.startSoln == Parameters::SmartStart) {
	    initialSoln() ; // nb: sorting of _q done in initialSoln()!!
        }
        else {
            // randomly shuffle the cells...
            for (n = 0 ; n < numCells ; n++) 
                _q.push_back(n) ;

            RandomUnsigned ru(0,_q.size()) ;
            for (n = 100*_q.size() ; n != 0 ; n--) {
                unsigned u = ru ;
		unsigned v = ru ;

                // debug... should rand upper lim be #, or #-1?
                if (u >= numCells)  u = numCells-1 ;
		if (v >= numCells)  v = numCells-1 ;
                
		unsigned tmp = _q[u] ; _q[u] = _q[v] ; _q[v] = tmp ;
            }
        }
    }

    // using _q workspace, build two sequences to interleave.
    // list composition depends on the splitType.

    vector<int> _a, _b ;
    if (_params.splitType == Parameters::RandomSplit) 
    {
        RandomUnsigned ru(0,_q.size()-1) ;

        // continue until no improvement.

        double bestCost = DBL_MAX; //, currCost = 0. ;
        for (unsigned pass = 0 ; pass != 20 ; pass++)
        {
            _a.clear() ; _b.clear() ;
	    unsigned split = ru ;
            if (split >= numCells-1) split = numCells-2 ;

	    for (n = 0 ; n < numCells ; n++) 
            {
	        if (n <= split) _a.push_back(_q[n]) ;
	        else            _b.push_back(_q[n]) ;
	    }

            double currCost = go(_a,_b) ; // do it!
	    if (currCost < bestCost-1.0e-8) 
	    {
	        bestCost=currCost;
                _problem.getSolnBuffer()[0].populated = true ;
	        for (n = 0 ; n != numCells ; n++) {soln[n].x = _curSoln[n].x;}
            }
	    else break;
        }
	return ;
    }
    else if (_params.splitType == Parameters::EvenSplit) 
    {
        double bestCost = DBL_MAX; //, currCost = 0. ;
        for (unsigned pass = 0 ; pass != 20 ; pass++)
        {
            _a.clear() ; _b.clear() ;
            unsigned split = numCells/2-1 ;
	    for (n = 0 ; n < numCells ; n++) {
	        if (n <= split) _a.push_back(_q[n]) ;
	        else            _b.push_back(_q[n]) ;
	    }

            double currCost = go(_a,_b) ; // do it!
	    if (currCost < bestCost-1.0e-8) 
            {
                bestCost=currCost;
                _problem.getSolnBuffer()[0].populated = true ;
	        for (n = 0 ; n != numCells ; n++) {soln[n].x = _curSoln[n].x;}
	    }
	    else break;
        }
	return ;
    }
    else if (_params.splitType == Parameters::AllSplit) 
    {
        vector<int> bestOrd(numCells) ;
	for (n = 0 ; n != numCells ; n++) bestOrd[n] = _q[n] ;

        double bestCost = DBL_MAX;
	for (unsigned pass = 0 ; pass != 20 ; pass++)
	{

	    // consider all splits of the order.
            double bestCostOverPass = DBL_MAX;
            for (unsigned split = 0 ; split != numCells ; split++)
	    {
	        _a.clear() ; _b.clear() ;
		for (n = 0 ; n != numCells ; n++)
		{
		    if (n <= split) _a.push_back(bestOrd[n]);
		    else            _b.push_back(bestOrd[n]);
		}
		double currCost = go(_a,_b);

		if (currCost < bestCostOverPass-1.0e-8)
		{
		    // nb: best soln during this pass; use to prep next.
                    bestCostOverPass = currCost;
		    for (n = 0 ; n != numCells ; n++) 
                    {
                        bestOrd[n]=_q[n];
                        _bestSolnOverPass[n].x = _curSoln[n].x;
                    }
		}
	    }
            if (bestCostOverPass<bestCost-1.0e-8)
	    {
	        bestCost=bestCostOverPass;
                _problem.getSolnBuffer()[0].populated = true ;
	        for (n = 0 ; n != numCells ; n++) 
                    soln[n].x = _bestSolnOverPass[n].x;
	    }
	    else break;
	}
    }
    else 
    {
        unsigned split = numCells/2-1 ; // default to even splitting.
	for (n = 0 ; n < numCells ; n++) {
	    if (n <= split) _a.push_back(_q[n]) ;
	    else            _b.push_back(_q[n]) ;
	}

        go(_a,_b) ; // do it!
	return ;
    }
}

double
DPSmallPlacer::go(vector<int>& _a, vector<int>& _b)
{
    // routine which does the actual DP.
//  const int numNets  = _problem.getNumNets() ;
    const int numCells = _problem.getNumCells() ;

    unsigned i, j; //, n;

    struct tableEntry {
        StripNetStacks* _stacks ;  
	int		_dir ;
    } ;

    tableEntry** _entries = new tableEntry* [_a.size()+1] ;
    for (i = 0 ; i <= _a.size() ; i++) {
        _entries[i] = new tableEntry [_b.size()+1] ;
        for (j = 0 ; j <= _b.size() ; j++) {
	    _entries[i][j]._dir    = -1 ;
            _entries[i][j]._stacks =  0 ;
        }
    }
    _entries[0][0]._dir    = -1 ;
    _entries[0][0]._stacks =  new StripNetStacks(_problem,_firstNets) ;

    for (i = 1 ; i <= _a.size() ; i++) {
        _entries[i][0]._dir = 0 ;
        _entries[i][0]._stacks = new StripNetStacks(_problem,_firstNets) ;
	_entries[i][0]._stacks->copy(*_entries[i-1][0]._stacks) ;
	_entries[i][0]._stacks->addNodeToLeft(_a[i-1]) ;
    }
    for (j = 1 ; j <= _b.size() ; j++) {
        _entries[0][j]._dir = 1 ;
        _entries[0][j]._stacks = new StripNetStacks(_problem,_firstNets) ;
	_entries[0][j]._stacks->copy(*_entries[0][j-1]._stacks) ;
	_entries[0][j]._stacks->addNodeToLeft(_b[j-1]) ;
    }

    for (i = 1 ; i <= _a.size() ; i++) {
        unsigned ai = _a[i-1] ;
        for (j = 1 ; j <= _b.size() ; j++) {
	    unsigned bj = _b[j-1] ;
	 
            StripNetStacks* temp1 = new StripNetStacks(_problem,_firstNets) ;
            StripNetStacks* temp2 = new StripNetStacks(_problem,_firstNets) ;

	    temp1->copy(*_entries[i-1][j]._stacks) ;
	    temp1->addNodeToLeft(ai) ;

	    temp2->copy(*_entries[i][j-1]._stacks) ;
	    temp2->addNodeToLeft(bj) ;

	    if (temp1->getCost() <= temp2->getCost()) {
	        _entries[i][j]._dir = 0 ;
	        _entries[i][j]._stacks = temp1 ;
		delete temp2 ;
	    }
	    else {
	        _entries[i][j]._dir = 1 ;
		_entries[i][j]._stacks = temp2 ;
		delete temp1 ;
	    }
	}
    }

    // trace back solution...

    vector<unsigned> order ; 
    order.reserve(numCells) ;

    int idx = numCells ;
    i = _a.size() ;
    j = _b.size() ;

    int dir = _entries[i][j]._dir ;
    for (; dir != -1 ; dir = _entries[i][j]._dir) {
        order[--idx] = ((dir == 0) ? (_a[i-1]) : (_b[j-1])) ;
	if (dir == 0)      i-- ;
	else if (dir == 1) j-- ;
	else ; // huh?

        _q[idx] = order[idx] ;
    }

    StripNetStacks finalStacks(_problem,_firstNets) ;

    _curWL = 0. ;
    for (idx = 0 ; idx < numCells ; idx++) {
        unsigned c = order[idx] ;

	_curSoln[c].x = finalStacks.getLeftEdge() ;
	_curWL = finalStacks.addNodeToLeft(c) ;
    }

    for (i = 0 ; i <= _a.size() ; i++) 
    {
        for (j = 0 ; j <= _b.size() ; j++)
        {
	    if (_entries[i][j]._stacks) 
	        delete _entries[i][j]._stacks ;
        }
	delete[] _entries[i];
    }
    delete[] _entries ;
     
    return _curWL ;
}

DPSmallPlacer::~DPSmallPlacer()
{
    if (_adjNets)    delete[] _adjNets ;
    if (_firstNets)  delete[] _firstNets ;
}

void DPSmallPlacer::buildConnectivity()
{
    const SmallPlacementNetlist& netlist = _problem.getNetlist() ;

    unsigned numCells = _problem.getNumCells() ;
    unsigned numNets  = _problem.getNumNets() ;
    unsigned numPins  = numNets * (numCells+1) ;

    _adjNets   = new EdgeAndOffset  [numPins] ;
    _firstNets = new EdgeAndOffset* [numCells+1] ;

    unsigned c, n, idx = 0 ;

        // nets (and offsets) adjacent to each node.
    for (c = 0 ; c < _problem.getNumCells() ; c++) 
    {
        _firstNets[c] = _adjNets + idx ;
        for (n = 0 ; n < _problem.getNumNets() ; n++)
        {
            double pinOffset = netlist.getPinOffset(c,n).x ;

            if(pinOffset != DBL_MAX)    //they are connected...
            {
                _adjNets[idx].edge = n ;
                _adjNets[idx].pinOffset = pinOffset ;
                idx++ ;
            }
        }
    }
    _firstNets[c] = _adjNets + idx ;
}

void DPSmallPlacer::initialSoln()
{

    const SmallPlacementNetlist& netlist = _problem.getNetlist() ;

    unsigned numCells = _problem.getNumCells() ;
    unsigned numNets  = _problem.getNumNets() ;
    unsigned numPins  = numNets * (numCells+1) ;

    double xMin = _problem.getRows()[0].xMin ;
    double xMax = _problem.getRows()[0].xMax ;

    unsigned* edgeDeg = new unsigned [numNets] ;
    unsigned* cellDeg = new unsigned [numCells] ;

    double* fixedTerminal = new double [numNets] ;

    unsigned*  adjCell   = new unsigned  [numPins+1] ;
    unsigned** firstCell = new unsigned* [numNets+1] ;

    unsigned*  adjNet    = new unsigned  [numPins+1] ;
    unsigned** firstNet  = new unsigned* [numCells+1] ;

    unsigned n, c, idx ;

        // determine edges adjacent to cells.
    for (idx = 0, n = 0 ; n < numNets ; n++) 
    {
        edgeDeg[n]   = 0 ;

        firstCell[n] = adjCell + idx ;
        for (c = 0 ; c < numCells ; c++) 
        {
            double pinOffset = netlist.getPinOffset(c,n).x ;
            if (pinOffset != DBL_MAX) 
            {
                adjCell[idx++] = c ; edgeDeg[n]++ ;
            }
        }

        const BBox& netBBox = _problem.getNetTerminalBBox(n) ;
        if (!netBBox.isEmpty()) {
            fixedTerminal[n] = (netBBox.xMin + netBBox.xMax) / 2. ;
            edgeDeg[n]++ ;
        }
        else 
            fixedTerminal[n] = 0. ;
    }
    firstCell[n] = adjCell + idx ;

    for (idx=0, c = 0 ; c < numCells ; c++) 
    {
        cellDeg[c] = 0 ;

        firstNet[c] = adjNet + idx ;
        for (n = 0 ; n < numNets ; n++) 
        {
            double pinOffset = netlist.getPinOffset(c,n).x ;
            if (pinOffset != DBL_MAX) 
            {
                adjNet[idx++] = n ; cellDeg[c]++ ;
            }
        }

        _curSoln[c].x = (xMin+xMax) / 2. ;
    }
    firstNet[c] = adjNet + idx ;

        // perform cog iterations.
    for (unsigned iter = 1 ; iter != 11 ; iter++)
    {
        for (unsigned curr = 0 ; curr < numCells ; curr++) // update cell i posn.
        {
            double cellPosn = 0. ;
            for (unsigned* i = firstNet[curr] ; i != firstNet[curr+1] ; i++)
            {
                n = *i ;

                double netPosn = fixedTerminal[n] ;
                for (unsigned* j = firstCell[n] ; j != firstCell[n+1] ; j++)
                {
                    c = *j ;
                    netPosn += (_curSoln[c].x + netlist.getPinOffset(c,n).x) ;
                }
                cellPosn += (netPosn / ((double) edgeDeg[n])) ;
            }
            if (cellDeg[curr] != 0)
                _curSoln[curr].x = cellPosn / ((double) cellDeg[curr]) ;
        }
    }

//for (c = 0 ; c < numCells ; c++) 
//    cout << " " << c << " " << _curSoln[c].x << endl ;

    for (c = 0 ; c < numCells ; c++) 
        _q.push_back(c) ;
    std::sort(_q.begin(), _q.end(), DPCompareNodesByInitPl(_curSoln));

//for (c = 0 ; c < numCells ; c++) 
//    cout << " " << c << " " << _q[c] << " " << _curSoln[_q[c]].x << endl ;

    delete[] adjNet  ; delete[] firstNet ;
    delete[] adjCell ; delete[] firstCell ;

    delete[] fixedTerminal ;
    delete[] edgeDeg ; delete[] cellDeg ;
}





