/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/









#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include <list.h>
#include "twoRowDPPlacer.h"

twoRowDPSmallPlacer::twoRowDPSmallPlacer(SmallPlacementProblem& problem,
		Parameters params)
	: BaseSmallPlacer(problem, params), 
	  _curSoln(problem.getNumCells())
{
    abkfatal(problem.getNumRows() == 2, "DP2D requires two rows") ;
    abkfatal(problem.getNumCells() >= 2, "DP2D requires >= 2 cells") ;

    unsigned bits_in_uns = sizeof(unsigned)*8 ;
    abkfatal(problem.getNumNets() < 2 * bits_in_uns, 
                            "DP2D can't deal with this many nets") ;


    Placement& soln    = _problem.getSolnBuffer()[0].placement ;
    //const unsigned numNets  = _problem.getNumNets() ;
    const unsigned numCells = _problem.getNumCells() ;

    buildConnectivity() ;


    // create cell list either by random shuffle, or using provided soln.

    unsigned n;
    _usingInitSoln = false;

    _q.reserve(numCells) ; // workspace to get init soln.

    _minRowViol[0] = DBL_MAX;
    _minRowViol[1] = DBL_MAX;   

    vector<vector<int> > _cellsInRow;
    _cellsInRow.resize(2);


    if(_problem.getSolnBuffer()[0].populated && _params.useInitSoln)
    //if(_params.useInitSoln)
    {
        //Push cells into row 0 until full, then row 1.
        for (n = 0 ; n < numCells ; n++) 
           _q.push_back(n);

        double left[2];
	left[0] = _problem.getRows()[0].xMin;
        left[1] = _problem.getRows()[1].xMin;
        unsigned row = 0;
        for (n = 0 ; n < numCells ; n++)
        {
	    double width = _problem.getCellWidth(_q[n]);
	    if (row == 0) 
	    {
                //nb: make sure at least one cell in row 0!
	        if (    (_cellsInRow[0].size() == 0) 
		     || (left[0]+width <= _problem.getRows()[0].xMax))
	            row = 0;
                else 
		    row = 1;
            }

            left[row] += width;

	    _cellsInRow[row].push_back(_q[n]);
        }

        //_minRowViol[0] = max(0.0,(left[0]-_problem.getRows()[0].xMax));
        //_minRowViol[1] = max(0.0,(left[1]-_problem.getRows()[1].xMax));
        _usingInitSoln = true;
    }
    else 
    {
        //If no initial solution, then use something random.
	//++++++++++++ can we be smarter here???
        if (_params.startSoln == Parameters::RandomStart)
        {
            // randomly shuffle the cells...
            for (n = 0 ; n < numCells ; n++) 
                _q.push_back(n);

            RandomUnsigned ru(0,_q.size()) ;
            for (n = 100*_q.size() ; n != 0 ; n--) {
                unsigned u = ru ;
                unsigned v = ru ;

                if (u >= numCells)  u = numCells-1 ;
	        if (v >= numCells)  v = numCells-1 ;
                
	        unsigned tmp = _q[u] ; _q[u] = _q[v] ; _q[v] = tmp ;
            }
        }
        else 
	{
	    for (n = 0 ; n < numCells ; n++)
	        _q.push_back(n);
	}

        double left[2];
	left[0] = _problem.getRows()[0].xMin;
        left[1] = _problem.getRows()[1].xMin;
 
        unsigned row = 0;
        for (n = 0 ; n < numCells ; n++)
        {
	    double width = _problem.getCellWidth(_q[n]);
	    if (row == 0) 
	    {
                //nb: make sure at least one cell in row 0!
	        if (    (_cellsInRow[0].size() == 0) 
		     || (left[0]+width <= _problem.getRows()[0].xMax))
	            row = 0;
                else 
		    row = 1;
            }
/*
	    else
	    {
	        if ( (left[1]+width <= _problem.getRows()[1].xMax))
	            row = 1;
                else if(left[0]+width <= _problem.getRows()[0].xMax)
		    row = 0;
	    }
*/
            left[row] += width;

	    _cellsInRow[row].push_back(_q[n]);
        }
        //_minRowViol[0] = max(0.0,(left[0]-_problem.getRows()[0].xMax));
        //_minRowViol[1] = max(0.0,(left[1]-_problem.getRows()[1].xMax));
    }
    //Determine cost of the initial solution.
    twoRowDPStacks initStk(_problem,_firstNets);

    unsigned r;
    for (r = 0 ; r <= 1 ; r++)
    {
        for (n = 0 ; n < _cellsInRow[r].size() ; n++)
        {
            int cellId = _cellsInRow[r][n];

	    soln[cellId].x = _curSoln[cellId].x = initStk.getLeftEdge(r);
	    soln[cellId].y = _curSoln[cellId].y = initStk.getBottomEdge(r);

	    initStk.addNode(cellId,r);
        }

    }
    _problem.getSolnBuffer()[0].populated = true;

    _bestCost      = initStk.getCost();
    _minRowViol[0] = max(0.0,(initStk.getLeftEdge(0)-_problem.getRows()[0].xMax));
    _minRowViol[1] = max(0.0,(initStk.getLeftEdge(1)-_problem.getRows()[1].xMax));


    //repeat until no improvement.  

    while (go(_cellsInRow) != DBL_MAX)
    {
        twoRowDPStacks stk(_problem,_firstNets);

        for (r = 0 ; r <= 1 ; r++)
        {
            for (n = 0 ; n < _cellsInRow[r].size() ; n++)
            {
                int cellId = _cellsInRow[r][n];

	        soln[cellId].x = _curSoln[cellId].x;
	        soln[cellId].y = _curSoln[cellId].y;

	        stk.addNode(cellId,r);
            }
        }
	_problem.getSolnBuffer()[0].populated = true;

	_bestCost = stk.getCost();
        _minRowViol[0] = max(0.0,(stk.getLeftEdge(0)-_problem.getRows()[0].xMax));
        _minRowViol[1] = max(0.0,(stk.getLeftEdge(1)-_problem.getRows()[1].xMax));
    }
}

double
twoRowDPSmallPlacer::go(vector<vector<int> >& _list)
{
    const unsigned numCells = _problem.getNumCells();

    double rightEdge[2];
    rightEdge[0] = _problem.getRows()[0].xMax;
    rightEdge[1] = _problem.getRows()[1].xMax;

    unsigned lst, row;

    // make the table.

    unsigned i,j,k,l;
    unsigned dim1 = _list[0].size()+1;
    unsigned dim2 = _list[0].size()+1;
    unsigned dim3 = _list[1].size()+1;
    unsigned dim4 = _list[1].size()+1;

    tableEntry ****entries ;

    entries = new tableEntry*** [dim1];
    for (i = 0 ; i != dim1 ; i++)
    {
        entries[i] = new tableEntry** [dim2];
        for (j = 0 ; j != dim2 ; j++)
	{
	    entries[i][j] = new tableEntry* [dim3];
	    for (k = 0 ; k != dim3 ; k++)
	    {
	        entries[i][j][k] = new tableEntry [dim4];
		for (l = 0 ; l != dim4 ; l++)
		{
                    vector<unsigned> t(4);
		    t[0] = i; t[1] = j; t[2] = k; t[3] = l;

		    entries[i][j][k][l].pt = coords(t);
		    entries[i][j][k][l].st = new twoRowDPStacks(_problem,_firstNets);

		    entries[i][j][k][l].added  = false;
		    entries[i][j][k][l].parent = 0;
		}
	    }
	}
    }



    //use DFS to search thru the solution space.

    double bestViol[2];
    bestViol[0]     = _minRowViol[0];
    bestViol[1]     = _minRowViol[1];
    double bestCost = _bestCost;

    tableEntry* bestEntry = 0;

    list<coords*> _myqueue;
    entries[0][0][0][0].added = true;
    _myqueue.push_back(&(entries[0][0][0][0].pt));
    while (_myqueue.size())
    {

	coords* head = _myqueue.front();
        _myqueue.pop_front();

	unsigned x1=head->x[0];
	unsigned x2=head->x[1];
	unsigned x3=head->x[2];
	unsigned x4=head->x[3];


        if (x1+x2+x3+x4 == numCells)
	{
	    //complete solution; test for quality.

	    twoRowDPStacks& stk = *(entries[x1][x2][x3][x4].st);

            double currCost = stk.getCost();
	    double currViol[2];
	    currViol[0] = max(0.0,stk.getLeftEdge(0)-rightEdge[0]);
	    currViol[1] = max(0.0,stk.getLeftEdge(0)-rightEdge[0]);

	    if (max(currViol[0],currViol[1]) <= max(bestViol[0],bestViol[1])) 
            {
	        //better or same violation.
		if (currCost < bestCost)
		{
		    //better cost; keep solution.

		    bestEntry = &entries[x1][x2][x3][x4];

		    bestCost = currCost;
		    bestViol[0] = currViol[0];
		    bestViol[1] = currViol[1];
		}
	    }
	}
	else 
	{
	    //generate the (possibly) four children of the current state.
	    //bound if possible.

	    for (lst = 0 ; lst <= 1 ; lst++)
	    {
	        unsigned cnt = ((lst==0) ? (x1+x2) : (x3+x4));
		if (cnt >= _list[lst].size())
		    continue; //current list exhausted.

                unsigned cellid = _list[lst][cnt]; //cell to add.

		for (row = 0 ; row <= 1 ; row++)
		{
		    //get child coordinates.
		    unsigned y1 = ((lst == 0 && row == 0) ? (x1+1) : x1);
		    unsigned y2 = ((lst == 0 && row == 1) ? (x2+1) : x2);
		    unsigned y3 = ((lst == 1 && row == 0) ? (x3+1) : x3);
		    unsigned y4 = ((lst == 1 && row == 1) ? (x4+1) : x4);

		    twoRowDPStacks tmpStk(_problem,_firstNets);

		    tmpStk.copy(*(entries[x1][x2][x3][x4].st));
		    tmpStk.addNode(cellid,row);

                    //bound away from crappy solutions.

		    double viol[2];
		    viol[0] = max(0.0,tmpStk.getLeftEdge(0)-rightEdge[0]);
		    viol[1] = max(0.0,tmpStk.getLeftEdge(1)-rightEdge[1]);

                    bool bnd = false;
		    if (max(viol[0],viol[1]) > max(bestViol[0],bestViol[1]))
		        bnd = true; //worse violation, so avoid.

                    if (_usingInitSoln)
		    {
		        if (tmpStk.getCost() >= bestCost)
			    bnd = true; //must improve cost if init soln provided.
		    }
		    else
		    {
		        if (max(viol[0],viol[1]) == max(bestViol[0],bestViol[1]))
			{
			    if (tmpStk.getCost() >= bestCost)
			        bnd = true; //no violation imp, and worst cost.
			}
		    }

                    if (bnd) 
                    {
		        //cout<<"bound away..."<<endl;
			continue;
		    }

                    /**************
	            cout<<"\taddCell: "<<cellid<<" "
		        <<"list: "<<lst<<" "
			<<"row: "<<row<<" "
			<<"cost: "<<tmpStk.getCost()
			<<endl;
                     **************/

                    if (!entries[y1][y2][y3][y4].added)
		    {
                        entries[y1][y2][y3][y4].added  = true;
                        entries[y1][y2][y3][y4].parent = &entries[x1][x2][x3][x4];
                        entries[y1][y2][y3][y4].st->copy(tmpStk);
			_myqueue.push_back(&(entries[y1][y2][y3][y4].pt));
		    }
		    else if (tmpStk.getCost() < entries[y1][y2][y3][y4].st->getCost())
		    {
                        entries[y1][y2][y3][y4].parent = &entries[x1][x2][x3][x4];
                        entries[y1][y2][y3][y4].st->copy(tmpStk);
		    }
		}
	    }
	}
    }

    if (bestEntry != 0)
    {
        //did find an improvement.  trace path to get assigment order.
        vector<tableEntry*> stk;

        tableEntry* curr = bestEntry;
        while (curr)
        {
            stk.push_back(curr);
	    curr = curr->parent;
        }

	vector<vector<unsigned> > sln;
	sln.resize(2);

        unsigned idx[2];
	idx[0] = idx[1] = 0;

	twoRowDPStacks fStk(_problem,_firstNets);

        unsigned i;
        for (i = stk.size() ; i > 0 ; )
        {
            i--;

	    curr = stk[i];

            /***
	    cout<<"...entry: "
	        <<"("
	        <<curr->pt.x[0]<<","<<curr->pt.x[1]
	        <<":"
	        <<curr->pt.x[2]<<","<<curr->pt.x[3]
	        <<")"
	        <<endl;
             ***/
         
	    if (i != 0)
	    {
	        lst = 0;
		row = 0;

	        if (stk[i]->pt.x[0] != stk[i-1]->pt.x[0])
		{
		    lst = 0; row = 0;
                }
	        else if (stk[i]->pt.x[1] != stk[i-1]->pt.x[1])
		{
		    lst = 0; row = 1;
                }
	        else if (stk[i]->pt.x[2] != stk[i-1]->pt.x[2])
		{
		    lst = 1; row = 0;
                }
	        else if (stk[i]->pt.x[3] != stk[i-1]->pt.x[3])
		{
		    lst = 1; row = 1;
                }

                unsigned cellId = _list[lst][idx[lst]++];

		sln[row].push_back(cellId);

		_curSoln[cellId].x = fStk.getLeftEdge(row);
		_curSoln[cellId].y = fStk.getBottomEdge(row);

		fStk.addNode(cellId,row);
	    }
        }

        bestCost = fStk.getCost();

	for (lst = 0 ; lst <= 1 ; lst++)
	{
	    _list[lst].clear();
	    for (i = 0 ; i < sln[lst].size() ; i++)
	        _list[lst].push_back(sln[lst][i]);
	}
    }
    else
    {
        bestCost = DBL_MAX;//nb: force a termination.
    }

    //free the DP table entries.
    for (i = 0 ; i != dim1 ; i++)
    {
        for (j = 0 ; j != dim2 ; j++)
	{
	    for (k = 0 ; k != dim3 ; k++)
	    {
	        for (l = 0 ; l != dim4 ; l++)
		{
		    delete entries[i][j][k][l].st;
		}
	        delete[] entries[i][j][k];
	    }
	    delete[] entries[i][j];
	}
        delete[] entries[i];
    }
    delete[] entries;

    return bestCost;
}

twoRowDPSmallPlacer::~twoRowDPSmallPlacer()
{
    if (_adjNets)    delete[] _adjNets ;
    if (_firstNets)  delete[] _firstNets ;
}

void twoRowDPSmallPlacer::buildConnectivity()
{
    const SmallPlacementNetlist& netlist = _problem.getNetlist() ;

    const unsigned numNets  = _problem.getNumNets() ;
    const unsigned numCells = _problem.getNumCells() ;

    unsigned numPins  = numNets * (numCells+1) ;

    _adjNets   = new EdgeAndOffset  [numPins] ;
    _firstNets = new EdgeAndOffset* [numCells+1] ;

    unsigned c, n, idx = 0 ;

        // nets (and offsets) adjacent to each node.
    for (c = 0 ; c < _problem.getNumCells() ; c++) 
    {
        _firstNets[c] = _adjNets + idx ;
        for (n = 0 ; n < _problem.getNumNets() ; n++)
        {
            double pinOffsetX = netlist.getPinOffset(c,n).x ;
            double pinOffsetY = netlist.getPinOffset(c,n).y ;

            if(pinOffsetX != DBL_MAX)    //they are connected...
            {
                _adjNets[idx].edge = n ;
                _adjNets[idx].pinOffsetX = pinOffsetX ;
                _adjNets[idx].pinOffsetY = pinOffsetY ;
                idx++ ;
            }
        }
    }
    _firstNets[c] = _adjNets + idx ;
}

void twoRowDPSmallPlacer::initialSoln()
{
    // what should we do for an initial soln?
    ;
}

void twoRowDPSmallPlacer::traceBack(tableEntry* ent)
{
    //trace partial solution.

    cout<<"traceback..."<<endl;

    vector<tableEntry*> stk;

    tableEntry* curr = ent;
    while (curr)
    {
        stk.push_back(curr);
	curr = curr->parent;
    }

    unsigned i;
    for (i = stk.size() ; i > 0 ; )
    {
        i--;

	curr = stk[i];

	cout<<"...entry: "
	    <<"("
	    <<curr->pt.x[0]<<","<<curr->pt.x[1]
	    <<":"
	    <<curr->pt.x[2]<<","<<curr->pt.x[3]
	    <<")"
	    <<endl;
    }
}



